# coding: utf-8
#  -*- Mode: Python; -*-                                              
#
# python test.py [local | system]
#  AUTHOR Jenna N. Shockley
#  AUTHOR Jamie A. Jennings

# TODO:
# - replace magic error code numbers with constants
# python test.py [local | system]
#
from __future__ import unicode_literals, print_function

import unittest
import sys, os, json
import rosie

# Notes
#
# (1) We use the librosie.so that is in the directory above this one,
#     i.e. "..", so we supply the librosiedir argument to
#     rosie.engine().  Normally, no argument to rosie.engine() is
#     needed.

if sys.version_info.major < 3:
    str23 = lambda s: str(s)
    bytes23 = lambda s: bytes(s)
else:
    str23 = lambda s: str(s, encoding='UTF-8')
    bytes23 = lambda s: bytes(s, encoding='UTF-8')

# -----------------------------------------------------------------------------
# Tests for Engine class
# -----------------------------------------------------------------------------

class RosieEngineTest(unittest.TestCase):

    def setUp(self):
        rosie.load(librosiedir, quiet=True)
        self.engine = rosie.engine()
        pass

    def tearDown(self):
        pass

    def testInit(self):
        assert(self.engine)
        path = rosie.librosie_path()
        assert(path)

    def testLoad(self):
        try:
            pkgname = self.engine.load(b'package x; foo = "foo"')
            self.assertTrue(pkgname == b"x")
        except RuntimeError:
            self.fail()

        try:
            b = self.engine.compile(b"x.foo")
            self.assertTrue(b.valid())
        except RuntimeError:
            self.fail()

        try:
            bb = self.engine.compile(b"[:digit:]+")
            self.assertTrue(bb.valid())
        except RuntimeError:
            self.fail()

        b2 = None
        try:
            b2 = self.engine.compile(b"[:foobar:]+")
            self.fail()
        except RuntimeError as e:
            self.assertTrue(not b2)
            self.assertTrue(str(e))
            #todo fix
            #self.assertEqual(str(e), "RPL error:\nb\'[{\"who\":\"compiler\",\"formatted\":\"Compile error\\n\\t[compiler]: unknown named charset: foobar\\n\\tin user input :1:3: [:foobar:]+\",\"message\":\"unknown named charset: foobar\"}]\'")

        try:
            b = None                       # trigger call to librosie to gc the compiled pattern
            b = self.engine.compile(b"[:digit:]+")
        except RuntimeError:
            self.fail()

        num_int = None
        try:
            num_int = self.engine.compile(b"num.int")
            self.fail()
        except RuntimeError as e:
            self.assertTrue(not num_int)
            self.assertTrue(str(e))
            #todo fix
            self.assertEqual(str(e), str(e))

        try:
            pkgname = self.engine.load(b'foo = "')
            self.fail()
        except RuntimeError as e:
            self.assertTrue(str(e))
            #todo fix
            self.assertEqual(str(e), str(e))

        engine2 = rosie.engine()
        self.assertTrue(engine2)
        self.assertTrue(engine2 != self.engine)
        engine2 = None          # triggers call to librosie to gc the engine

    def testConfig(self):
        cfg = self.engine.config()
        self.assertTrue(cfg)
        array = cfg[0]
        for entry in array:
            self.assertTrue(type(entry) is dict)
            self.assertTrue(entry['name'])
            self.assertTrue(entry['description'])
            if not entry['value']:
                self.assertTrue(entry['name'] == 'ROSIE_COMMAND')
        array = cfg[1]
        rpl_version = None
        libpath = None
        for entry in array:
            if entry['name']=='RPL_VERSION':
                rpl_version = entry['value']
            if entry['name']=='ROSIE_LIBPATH':
                libpath = entry['value']
        if sys.version_info.major < 3:
            self.assertTrue(type(rpl_version) is unicode)
            self.assertTrue(type(libpath) is unicode)
        else:
            self.assertTrue(type(rpl_version) is str)
            self.assertTrue(type(libpath) is str)

    def testLibpath(self):
        path = self.engine.libpath()
        self.assertIsInstance(path, bytes)
        newpath = b"foo bar baz"
        self.engine.libpath(newpath)
        testpath = self.engine.libpath()
        self.assertIsInstance(testpath, bytes)
        self.assertTrue(testpath == newpath)
        
    def testAllocLimit(self):
        limit, usage = self.engine.alloc_limit()
        self.assertIsInstance(limit, int)
        self.assertTrue(limit == 0)
        limit, usage = self.engine.alloc_limit(0)
        limit, usage = self.engine.alloc_limit()
        self.assertTrue(limit == 0)
        limit, usage = self.engine.alloc_limit(8199)
        self.assertTrue(limit == 8199)
        with self.assertRaises(ValueError):
            limit, usage = self.engine.alloc_limit(8191) # too low
        limit, usage = self.engine.alloc_limit()
        self.assertTrue(limit == 8199)
            
    def testImport(self):
        try:
            pkgname = self.engine.import_package(b'net')
            self.assertTrue(pkgname == b'net')
        except RuntimeError:
            self.fail()

        try:
            pkgname = self.engine.import_package(b'net', b'foobar')
            self.assertTrue(pkgname == b'net') # actual name inside the package
        except RuntimeError:
            self.fail()

        net_any = None
        try:
            net_any = self.engine.compile(b"net.any")
            self.assertTrue(net_any)
        except RuntimeError:
            self.fail()

        try:
            foobar_any = self.engine.compile(b"foobar.any")
            self.assertTrue(foobar_any)
        except RuntimeError:
            self.fail()
        
        m = net_any.match(b"1.2.3.4", 1, encoder = b"color")
        self.assertTrue(m)
        m = net_any.match(b"Hello, world!", 1, encoder = b"color")
        self.assertTrue(not m)

        try:
            pkgname = self.engine.import_package(b'THISPACKAGEDOESNOTEXIST')
        except RuntimeError as e:
            self.assertTrue(e)
    
    def testLoadFile(self):
        try:
            pkgname = self.engine.loadfile(b'test.rpl')
            self.assertTrue(pkgname == bytes(b'test'))
        except RuntimeError:
            self.fail()

    def testMatch(self):
        # test short match
        try:
            m = self.engine.match(b"[:digit:]+", b"321")
            self.assertTrue(m)
            self.assertTrue(m.start() == 1)     # match started at char 2
            self.assertTrue(m.end() == 4)
            self.assertTrue(m.group() == "321")
            self.assertTrue(m.abend() == False)
        except:
            self.fail()

        # test no match
        try:
            m = self.engine.match(b"[:digit:]+", b"xyz")
            self.assertTrue(m == None)
        except:
            self.fail()

        # test long match
        inp = b"889900112233445566778899100101102103104105106107108109110xyz"
        linp = len(inp)

        try:
            m = self.engine.match(b"[:digit:]+", inp)
            self.assertTrue(m)
            self.assertTrue(m.start() == 1)
            self.assertTrue(m.end() == linp-3+1) # due to the "xyz" at the end
            self.assertTrue(m.group() == str23(inp[0:-3]))
            self.assertTrue(m.abend() == False)
        except:
            self.fail()

    def testFullMatch(self):
        try:
            pkgname = self.engine.load(b'package year; d = [0-9]')
        except:
            self.fail()

        m = self.engine.fullmatch(b'year.d{4}', b"1998")
        self.assertTrue(m)
        self.assertEqual(m.group(), "1998")

        m = self.engine.fullmatch(b'year.d{4}', b"1998 Year")
        self.assertTrue(m == None)

    def testTrace(self):
        pkgname = self.engine.import_package(b'net')
        self.assertTrue(pkgname)

        dict = self.engine.trace(b'net.fqdn_strict', b"a.b.c", encoder=b"condensed")
        self.assertTrue(dict['matched'] == True)
        self.assertTrue(dict['trace'])
        self.assertTrue(len(dict['trace']) > 0)

        dict = self.engine.trace(b"net.fqdn_strict", b"a.b.c", encoder=b"condensed")
        self.assertTrue(dict['matched'] == True)
        self.assertTrue(dict['trace'])
        self.assertTrue(len(dict['trace']) > 0)

        dict = self.engine.trace(b'net.fqdn_strict', b"a.b.c", encoder=b"full")
        self.assertTrue(dict['matched'] == True)
        self.assertTrue(dict['trace'])
        self.assertTrue(len(dict['trace']) > 0)
        self.assertTrue(dict['trace'].find(b'Matched 5 chars') != -1)

        try:
            dict = self.engine.trace(b'net.fqdn_strict', b"a.b.c", encoder=b"no_such_trace_style")
            # force a failure in case the previous line does not throw an exception:
            self.assertTrue(False)
        except ValueError as e:
            self.assertTrue(repr(e).find('invalid trace style') != -1)

        trace_object = self.engine.trace(b'net.fqdn_strict', b"a.b.c")
        self.assertTrue(trace_object.trace_value())
        self.assertTrue(len(trace_object.trace_value()) > 0)
        self.assertTrue('match' in trace_object.trace_value())
        self.assertTrue(trace_object.trace_value()['match'])
        self.assertTrue('nextpos' in trace_object.trace_value())
        self.assertTrue(trace_object.trace_value()['nextpos'] == 6)
    
    def testSearch(self):
        try:
            pkgname = self.engine.load(b'package year; d = [0-9]')
        except:
            self.fail()

        m = self.engine.search('year.d{4}', "info 1998 23 2006 time 1876")
        self.assertEqual(m.group(), "1998")

    def testFindall(self):
        try:
            pkgname = self.engine.load(b'package year; d = [0-9]')
        except:
            self.fail()

        matches = self.engine.findall('year.d{4}', "info 1998 23 2006 time 1876")
        self.assertTrue("1998" in matches)
        self.assertTrue("2006" in matches)
        self.assertTrue("1876" in matches)
        self.assertTrue("23" not in matches)

    def testSub(self):
        self.assertEqual(self.engine.sub('[0-9]{4}', 'time', 'the year of 1956 and 1957'), 'the year of time and time')

    def testSubn(self):
        self.assertEqual(self.engine.subn('[0-9]{4}', 'time', 'the year of 1956 and 1957'), ('the year of time and time', 2))

    def testReadRCFile(self):
        if testdir:
            options, messages = self.engine.read_rcfile(bytes23(os.path.join(testdir, "rcfile1")))
            self.assertIsInstance(options, list)
            self.assertTrue(messages is None)
            options, messages = self.engine.read_rcfile(bytes23(os.path.join(testdir, "rcfile2")))
            self.assertTrue(messages[0].find("Syntax errors in rcfile") != -1)
            self.assertTrue(options is False)
            options, messages = self.engine.read_rcfile(b"This file does not exist")
            self.assertTrue(options is None)
            self.assertTrue(messages[0].find("Could not open rcfile") != -1)
            
    def testExecuteRCFile(self):
        result, messages = self.engine.execute_rcfile(b"This file does not exist")
        self.assertTrue(result is None)
        if testdir:
            result, messages = self.engine.execute_rcfile(bytes23(os.path.join(testdir, "rcfile1")))
            self.assertFalse(result)
            self.assertTrue(messages[0].find("Failed to load another-file") != -1)
            result, messages = self.engine.execute_rcfile(bytes23(os.path.join(testdir, "rcfile2")))
            self.assertFalse(result)
            self.assertTrue(messages[0].find("Syntax errors in rcfile") != -1)
            result, messages = self.engine.execute_rcfile(bytes23(os.path.join(testdir, "rcfile3")))
            self.assertFalse(result)
            self.assertTrue(messages[0].find("Failed to load nofile_mod1.rpl") != -1)
            result, messages = self.engine.execute_rcfile(bytes23(os.path.join(testdir, "rcfile6")))
            self.assertTrue(result)
            self.assertTrue(messages is None)

    def testDeps(self):
        # Expression
        result, messages = self.engine.expression_deps(b"A")
        self.assertFalse(result is None)
        self.assertTrue(messages is None)
        pt = json.loads(result)
        self.assertTrue(len(pt) == 0)
        result, messages = self.engine.expression_deps(b'A / "hello" / B.c [:digit:]+ p.mac:#hi')
        def check_deps(result, messages, index=None):
            self.assertFalse(result is None)
            self.assertTrue(messages is None)
            pt = json.loads(result)
            if index: pt = pt[index]
            self.assertTrue(len(pt) == 2)
            self.assertTrue(pt[0] == 'B')
            self.assertTrue(pt[1] == 'p')
        check_deps(result, messages)
        result, messages = self.engine.expression_deps(b"A // B.c") # syntax error
        self.assertTrue(result is None)
        self.assertFalse(messages is None)
        # Block
        result, messages = self.engine.block_deps(b"x = A / B.c; y=[:alpha:] p.mac:#tagname")
        check_deps(result, messages, 'implicit')
        result, messages = self.engine.block_deps(b"import F as G, H; x = A / B.c; y=[:alpha:] p.mac:#tagname")
        check_deps(result, messages, 'implicit')
        pt = json.loads(result)
        pt = pt['explicit']
        self.assertTrue(len(pt) == 2)
        self.assertTrue('as_name' in pt[0])
        self.assertTrue(pt[0]['as_name'] == 'G')
        self.assertTrue('importpath' in pt[0])
        self.assertTrue(pt[0]['importpath'] == 'F')
        self.assertTrue(not 'as_name' in pt[1])
        self.assertTrue('importpath' in pt[1])
        self.assertTrue(pt[1]['importpath'] == 'H')
        result, messages = self.engine.block_deps(b" = A / B.c; y=[:alpha:]") # syntax error
        self.assertTrue(result is None)
        self.assertFalse(messages is None)

    def testRefs(self):
        # Expression
        result, messages = self.engine.expression_refs(b"A")
        self.assertFalse(result is None)
        self.assertTrue(messages is None)
        pt = json.loads(result)
        self.assertTrue(len(pt) == 1)
        self.assertTrue('ref' in pt[0])
        self.assertTrue('localname' in pt[0]['ref'])
        self.assertTrue(pt[0]['ref']['localname'] == "A")
        result, messages = self.engine.expression_refs(b'A / "hello" / B.c [:digit:]+ mac:#hi')
        def check_refs(result, messages):
            self.assertFalse(result is None)
            self.assertTrue(messages is None)
            pt = json.loads(result)
            self.assertTrue(len(pt) == 3)
            self.assertTrue('ref' in pt[0])
            self.assertTrue(pt[0]['ref']['localname'] == "A")
            self.assertTrue('ref' in pt[1])
            self.assertTrue(pt[1]['ref']['packagename'] == "B")
            self.assertTrue(pt[1]['ref']['localname'] == "c")
            self.assertTrue('ref' in pt[2])
            self.assertTrue(pt[2]['ref']['localname'] == "mac")
        check_refs(result, messages)
        result, messages = self.engine.expression_refs(b"A // B.c") # syntax error
        self.assertTrue(result is None)
        self.assertFalse(messages is None)
        # Block
        result, messages = self.engine.block_refs(b"x = A / B.c; y=[:alpha:] mac:#tagname")
        check_refs(result, messages)
        result, messages = self.engine.block_refs(b" = A / B.c; y=[:alpha:]") # syntax error
        self.assertTrue(result is None)
        self.assertFalse(messages is None)

    def testParse(self):
        # Parse expression
        result, messages = self.engine.parse_expression(b"A")
        self.assertFalse(result is None)
        self.assertTrue(messages is None)
        pt = json.loads(result)
        self.assertTrue('ref' in pt)
        self.assertTrue('localname' in pt['ref'])
        self.assertTrue(pt['ref']['localname'] == "A")
        result, messages = self.engine.parse_expression(b"A / B.c")
        self.assertFalse(result is None)
        self.assertTrue(messages is None)
        pt = json.loads(result)
        self.assertTrue('choice' in pt)
        self.assertTrue('exps' in pt['choice'])
        second_exp = pt['choice']['exps'][1]
        self.assertTrue('ref' in second_exp)
        self.assertTrue(second_exp['ref']['packagename'] == "B")
        self.assertTrue(second_exp['ref']['localname'] == "c")
        result, messages = self.engine.parse_expression(b"A // B.c") # syntax error
        self.assertTrue(result is None)
        self.assertFalse(messages is None)
        # Parse block
        result, messages = self.engine.parse_block(b"x = A / B.c; y=[:alpha:]")
        self.assertFalse(result is None)
        self.assertTrue(messages is None)
        pt = json.loads(result)
        self.assertTrue('block' in pt)
        self.assertTrue('stmts' in pt['block'])
        binding1 = pt['block']['stmts'][0]
        binding2 = pt['block']['stmts'][1]
        def check_binding(b, boundname):
            self.assertTrue('binding' in b)
            self.assertTrue('ref' in b['binding'])
            self.assertTrue('ref' in b['binding']['ref'])
            self.assertTrue('localname' in b['binding']['ref']['ref'])
            self.assertTrue(b['binding']['ref']['ref']['localname'] == boundname)
        check_binding(binding1, 'x')
        check_binding(binding2, 'y')
        result, messages = self.engine.parse_block(b" = A / B.c; y=[:alpha:]") # syntax error
        self.assertTrue(result is None)
        self.assertFalse(messages is None)

# -----------------------------------------------------------------------------
# Tests for RPLX class
# -----------------------------------------------------------------------------

class RosieRPLXTest(unittest.TestCase):

    engine = None
    
    def setUp(self):
        rosie.load(librosiedir, quiet=True)
        self.engine = rosie.engine()

    def tearDown(self):
        pass

    def test_match_object(self):
        b = None
        try:
            pkgname = self.engine.load(b'package email; x1 = [a-z]+; x2 = [a-z]+; x3 = [a-z]+')
            b = self.engine.compile(b'email.x1[@]email.x2[.]email.x3')
            self.assertTrue(b.valid())
        except:
            self.fail()

        m = b.match(b"user@ncsu.edu")
        self.assertTrue(m)
        self.assertTrue(m.start() == 1)     # match started at char 1
        self.assertEqual(m.end(), 14)
        self.assertEqual(m.group(), "user@ncsu.edu")
        self.assertEqual(m.groupdict(), {'*': 'user@ncsu.edu', 'x1': 'user', 'x2': 'ncsu', 'x3': 'edu'})
        self.assertEqual(m.group('x1', 'x2'), ("user", "ncsu"))
        self.assertEqual(m.group('x1'), "user")
        self.assertEqual(m['x1'], "user")
        self.assertEqual(m.start('x1'), 1)
        self.assertEqual(m.subs('x1'), ['user'])
        self.assertEqual(m.start(1), 1)
        self.assertEqual(m.end('x1'), 5)
        self.assertEqual(m.end(1), 5)
        self.assertEqual(m.span('x1'), (1,5))
        self.assertEqual(m.lastindex(), 3)
        self.assertEqual(m.lastgroup(), 'x3')
        self.assertEqual(m.group(0), "user@ncsu.edu")
        self.assertEqual(m.group(1), "user")
        self.assertEqual(m.group(2), "ncsu")
        self.assertEqual(m[2], "ncsu")
        self.assertEqual(m.group(3), "edu")
        self.assertEqual(m.group(1,2,3), ('user', 'ncsu', 'edu'))
        self.assertTrue(m.abend() == False)

        self.assertEqual(m.expand(r"hello \1 is done"), "hello user is done")
        self.assertEqual(m.expand(r'\1 is done'), "user is done")
        self.assertEqual(m.expand(r'username: \1 email: \2 end: \3'), "username: user email: ncsu end: edu")
        self.assertEqual(m.expand(r'username: \g<1> email: \g<2> end: \g<3>'), "username: user email: ncsu end: edu")
        self.assertEqual(m.expand(r'username: \g<x1> email: \g<x2> end: \g<x3>'), "username: user email: ncsu end: edu")

    def testSub(self):
        b = None
        b = self.engine.compile('[0-9]{4}')
        self.assertEqual(b.sub('time', 'the year of 1956 and 1957'), 'the year of time and time')

    def testSubn(self):
        b = None
        b = self.engine.compile('[0-9]{4}')
        self.assertEqual(b.subn('time', 'the year of 1956 and 1957'), ('the year of time and time', 2))

    def testFindIter(self):
        b = None
        b = self.engine.compile('[0-9]{4}')
        iter = b.finditer('the year of 1956 and 1957 and 1988', 1)
        li = [s for s in iter]
        self.assertEqual(len(li), 3)
        self.assertEqual(li[0], "1956")
        self.assertEqual(li[1], "1957")
        self.assertEqual(li[2], "1988")

    def testFullMatch(self):
        b = None
        try:
            pkgname = self.engine.load(b'package year; d = [0-9]')
            b = self.engine.compile(b'year.d{4}')
            self.assertTrue(b.valid())
        except:
            self.fail()

        m = b.fullmatch(b"1998", 1)
        self.assertTrue(m)
        self.assertEqual(m.group(), "1998")

        m = b.fullmatch(b"1998 Year", 1)
        self.assertTrue(m == None)

        m = b.fullmatch(b"1998 Year", 1, 5)
        self.assertTrue(m)
        self.assertEqual(m.group(), "1998")
        

    def testMatch(self):
        b = None
        try:
            b = self.engine.compile(b"[:digit:]+")
            self.assertTrue(b.valid())
        except:
            self.fail()

        # test short match
        try:
            m = b.match(b"321", 2)
            self.assertTrue(m)
            self.assertTrue(m.start() == 2)     # match started at char 2
            self.assertTrue(m.end() == 4)
            self.assertTrue(m.group() == "21")
            self.assertTrue(m.abend() == False)
        except:
            self.fail()

        # test no match
        try:
            m = b.match(b"xyz", 1)
            self.assertTrue(m == None)
        except:
            self.fail()

        # test long match
        inp = b"889900112233445566778899100101102103104105106107108109110xyz"
        linp = len(inp)

        try:
            m = b.match(inp, 1)
            self.assertTrue(m)
            self.assertTrue(m.start() == 1)
            self.assertTrue(m.end() == linp-3+1) # due to the "xyz" at the end
            self.assertTrue(m.group() == str23(inp[0:-3]))
            self.assertTrue(m.abend() == False)
        except:
            self.fail()

        try:
            m = b.match(inp, 10)
            self.assertTrue(m)
            self.assertTrue(m.start() == 10)
            self.assertTrue(m.end() == linp-3+1) # due to the "xyz" at the end
            self.assertTrue(m.group() == str23(inp[9:-3]))
            self.assertTrue(m.abend() == False)
        except:
            self.fail()

        #test other encoders
        try:
            m = b.match(inp, 1, encoder=b"line")
            self.assertTrue(m['match'])
            self.assertTrue(m['abend'] == False)
        except:
            self.fail()

        try:
            m = b.match(inp, 1, encoder=b"bool")
            self.assertTrue(m['match'])
            self.assertTrue(m['abend'] == False)
        except:
            self.fail()

        try:
            m = b.match(inp, 1, encoder=b"color")
            self.assertTrue(m)
            # only checking the first two chars, looking for the start of
            # ANSI color sequence
            self.assertTrue(str23(m['match'])[0] == '\x1B')
            self.assertTrue(str23(m['match'])[1] == '[')
            self.assertTrue(m['abend'] == False)
        except:
            self.fail()

    def testTrace(self):
        pkgname = self.engine.import_package(b'net')
        self.assertTrue(pkgname)
        self.net_any = self.engine.compile(b'net.any')
        self.assertTrue(self.net_any)

        dict = self.net_any.trace(b"a.b.c", 1, encoder=b"condensed")
        self.assertTrue(dict['matched'] == True)
        self.assertTrue(dict['trace'])
        self.assertTrue(len(dict['trace']) > 0)

        net_ip = self.engine.compile(b"net.ip")
        self.assertTrue(net_ip)
        dict = net_ip.trace(b"a.b.c", 1, encoder=b"condensed")
        self.assertTrue(dict['matched'] == False)
        self.assertTrue(dict['trace'])
        self.assertTrue(len(dict['trace']) > 0)

        dict = self.net_any.trace(b"a.b.c", 1, encoder=b"full")
        self.assertTrue(dict['matched'] == True)
        self.assertTrue(dict['trace'])
        self.assertTrue(len(dict['trace']) > 0)
        self.assertTrue(dict['trace'].find(b'Matched 5 chars') != -1)

        try:
            dict = self.net_any.trace(b"a.b.c", 1, encoder=b"no_such_trace_style")
            self.assertTrue(False)
        except ValueError as e:
            self.assertTrue(repr(e).find('invalid trace style') != -1)

        trace_object = self.net_any.trace(b"a.b.c", 1)
        self.assertTrue(trace_object.trace_value())
        self.assertTrue(len(trace_object.trace_value()) > 0)
        self.assertTrue('match' in trace_object.trace_value())
        self.assertTrue(trace_object.trace_value()['match'])
        self.assertTrue('nextpos' in trace_object.trace_value())
        self.assertTrue(trace_object.trace_value()['nextpos'] == 6)

    def testSearch(self):
        b = None
        try:
            pkgname = self.engine.load(b'package year; d = [0-9]')
            b = self.engine.compile('year.d{4}')
            self.assertTrue(b.valid())
        except:
            self.fail()

        m = b.search("info 1998 23 2006 time 1876", 1)
        self.assertEqual(m.group(), "1998")

    def testFindall(self):
        b = None
        try:
            pkgname = self.engine.load(b'package year; d = [0-9]')
            b = self.engine.compile('year.d{4}')
            self.assertTrue(b.valid())
        except:
            self.fail()

        matches = b.findall("info 1998 23 2006 time 1876", 1)
        self.assertTrue("1998" in matches)
        self.assertTrue("2006" in matches)
        self.assertTrue("1876" in matches)
        self.assertTrue("23" not in matches)



# -----------------------------------------------------------------------------

if __name__ == '__main__':
    if len(sys.argv) != 2:
        sys.exit("Error: missing command-line parameter specifying 'local' or 'system' test")
    if sys.argv[1]=='local':
        librosiedir = '//../binaries'
        print("Loading librosie from ", librosiedir[2:])
        testdir = os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))), '../../test')
    elif sys.argv[1]=='system':
        librosiedir = rosie.librosie_system()
        print("Loading librosie from system library path")
        testdir = None
    else:
        sys.exit("Error: invalid command-line parameter (must be 'local' or 'system')")
    print("Running tests using " + sys.argv[1] + " rosie installation")
    del sys.argv[1:]
    unittest.main()
    



#  -*- Mode: Python; -*-                                                   
#  -*- coding: utf-8; -*-
# 
#  decode.py
# 
#  © Copyright Jamie A. Jennings 2019.
#  LICENSE: MIT License (https://opensource.org/licenses/mit-license.html)
#  AUTHOR: Jamie A. Jennings

from __future__ import unicode_literals, print_function
import sys
assert( sys.version_info.major == 3 )


def decode_short(b, i):
    n = b[i] + (b[i+1]<<8)
    if n > 32767:
        n = n - 65536
    return n, i+2

def decode_int(b, i):
    n = b[i] + (b[i+1]<<8) + (b[i+2]<<16) + (b[i+3]<<24)
    if n > 2147483647:
        n = n - 4294967296
    return n, i+4

def decode(bytestring, i=0, start_position=None):
    if start_position is None:
        start_position, i = decode_int(bytestring, i)
    if start_position > 0:
        raise ValueError("Invalid capture start at position {}".format(i-4))
    else:
        start_position = - start_position
    typelen, i = decode_short(bytestring, i)
    data = None
    after_typename = i + (typelen if typelen >= 0 else -typelen)
    typename = bytestring[i:after_typename]
    i = after_typename
    if typelen < 0:
        # constant capture (user-provided string)
        typelen = - typelen
        datalen, i = decode_short(bytestring, i)
        assert( datalen >= 0 )
        data = bytestring[i:i+datalen]
        i = i + datalen
    # else: a regular capture, so the data is a substring of the input
    # and is not included here.
    # next, we peek ahead to see if there is a sub-match
    subs = list()
    next_position, _ = decode_int(bytestring, i) # peek
    while next_position < 0:
        # it's the start position of a nested capture
        sub, i = decode(bytestring, i+4, next_position)
        subs.append(sub)
        next_position, _ = decode_int(bytestring, i) # peek
    i = i + 4
    match = {'type': typename,
             'data': data,
             'subs': subs,
             's': start_position,
             'e': next_position} 
    return match, i




# coding: utf-8
#  -*- Mode: Python; -*-                                              
# 
#  rosie.py     An interface to librosie from Python 2.7 and 3.6
# 
#  © Copyright IBM Corporation 2016, 2017, 2018.
#  LICENSE: MIT License (https://opensource.org/licenses/mit-license.html)
#  AUTHOR: Jenna N. Shockley
#  AUTHOR: Jamie A. Jennings

# Development environment:
#   Mac OS X Sierra (10.13.3)
#   Python 2.7.10 (distributed with OS X)
#   Python 3.6.5 (installed via brew on OS X)
#   cffi-1.11.4
# 

from __future__ import unicode_literals, print_function

import sys, json

from . import internal
from . import adapt23

# -----------------------------------------------------------------------------
# Rosie-specific

def librosie_path():
    return internal._librosie_path

def librosie_system():
    return internal.librosie_system

def librosie_local():
    return internal.librosie_local

def load(path = None, **kwargs):
    return internal.load(path, kwargs)

# -----------------------------------------------------------------------------
# re Flags
'''class RegexFlag(enum.IntFlag):
    ASCII = RE_FLAG_ASCII # assume ascii "locale"
    IGNORECASE = RE_FLAG_IGNORECASE # ignore case
    LOCALE = RE_FLAG_LOCALE # assume current 8-bit locale
    UNICODE = RE_FLAG_UNICODE # assume unicode "locale"
    MULTILINE = RE_FLAG_MULTILINE # make anchors look for newline
    DOTALL = RE_FLAG_DOTALL # make dot match newline
    VERBOSE = RE_FLAG_VERBOSE # ignore whitespace and comments
    A = ASCII
    I = IGNORECASE
    L = LOCALE
    U = UNICODE
    M = MULTILINE
    S = DOTALL
    X = VERBOSE'''

# -----------------------------------------------------------------------------
# Rosie Engine

class engine (object):

    '''
    A Rosie pattern matching engine is used to load/import RPL code
    (patterns) and to do matching.  Create as many engines as you need.
    '''

    def __init__(self):
        self._engine = internal.engine()
    
    # -----------------------------------------------------------------------------
    # Compile an expression
    # -----------------------------------------------------------------------------

    def compile(self, exp):
        pat, errs = self._engine.compile(adapt23.bytes23(exp))
        if not pat:
            raise_rosie_error(errs)
        return rplx(self, exp, pat)


    # -----------------------------------------------------------------------------
    # Basic methods
    # included in python re
    # -----------------------------------------------------------------------------

    def match(self, pattern, string, **kwargs):
        """Try to apply the pattern at the start of the string, returning
        a match object, or None if no match was found."""
        #        errs = None
        pattern = self.compile(pattern)
        return pattern.match(string, **kwargs)
    
    def fullmatch(self, pattern, input, **kwargs):
        #        errs = None
        pattern = self.compile(pattern)
        return pattern.fullmatch(input, **kwargs)

    def trace(self, pattern, input, **kwargs):
        #        errs = None
        pattern = self.compile(pattern)
        return pattern.trace(input, **kwargs)

    def search(self, pattern, string, **kwargs):
        #        errs = None
        pattern = self.compile(pattern)
        return pattern.search(string, **kwargs)

    def findall(self, pattern, string, **kwargs):
        #        errs = None
        pattern = self.compile(pattern)
        return pattern.findall(string, **kwargs)

    def finditer(self, pattern, string, **kwargs):
        #        errs = None
        pattern = self.compile(pattern)
        return pattern.finditer(string, **kwargs)

    def sub(self, pattern, repl, string,**kwargs):
        #        errs = None
        pattern = self.compile(pattern)
        return pattern.sub(repl, string, **kwargs)

    def subn(self, pattern, repl, string,**kwargs):
        #        errs = None
        pattern = self.compile(pattern)
        return pattern.subn(repl, string, **kwargs)

    def purge(self):
        raise NotImplementedError("Purge is an 're' function that is not applicable to RPL.")


    # -----------------------------------------------------------------------------
    # Functions for loading statements/blocks/packages into an engine
    # this functionality not included in python re
    # -----------------------------------------------------------------------------

    def load(self, src):
        ok, pkgname, errs = self._engine.load(adapt23.bytes23(src))
        if not ok or errs:
            raise_rosie_error(errs)
        return pkgname

    def loadfile(self, filename):
        ok, pkgname, errs = self._engine.loadfile(adapt23.bytes23(filename))
        if not ok or errs:
            raise_rosie_error(errs)
        return pkgname

    def import_package(self, pkgname, as_name=None):
        ok, actual_pkgname, errs = self._engine.import_pkg(adapt23.bytes23(pkgname),
                                                           as_name and adapt23.bytes23(as_name))
        if not ok or errs:
            raise_rosie_error(errs)
        return actual_pkgname


    # -----------------------------------------------------------------------------
    # Functions for reading and processing rcfile (init file) contents
    # this functionality not included in python re
    # -----------------------------------------------------------------------------
    
    def read_rcfile(self, filename=None):
        return self._engine.read_rcfile(filename)

    def execute_rcfile(self, filename=None):
        return self._engine.execute_rcfile(filename)


    # -----------------------------------------------------------------------------
    # Functions that return a parse tree or fragments of one
    # this functionality not included in python re
    # -----------------------------------------------------------------------------

    def parse_expression(self, exp):
        return self._engine.parse_expression(exp)
        
    def parse_block(self, block):
        return self._engine.parse_block(block)
        
    def expression_refs(self, exp):
        return self._engine.expression_refs(exp)
        
    def block_refs(self, block):
        return self._engine.block_refs(block)
        
    def expression_deps(self, exp):
        return self._engine.expression_deps(exp)
        
    def block_deps(self, block):
        return self._engine.block_deps(block)

    # -----------------------------------------------------------------------------
    # Functions for reading and modifying various engine settings
    # this functionality not included in python re
    # -----------------------------------------------------------------------------

    def config(self):
        return json.loads(self._engine.config())

    def libpath(self, libpath=None):
        return self. _engine.libpath(libpath)

    def alloc_limit(self, newlimit=None):
        return self._engine.alloc_limit(newlimit)

    def __del__(self):
        self._engine = None


def raise_rosie_error(errs):
    raise RuntimeError('RPL error:\n{}'.format(errs))

def raise_halt_error():
    raise RuntimeError('Matching aborted due to pattern containing halt pattern')

class rplx(object):  #object for a compiled pattern  

    def __init__(self, engine, exp, internal_rplx):
        self.pattern = exp                                  # bytes or str
        self._internal_rplx = internal_rplx
        self.engine = engine
    
    def valid(self):
        return self._internal_rplx.valid
    
    # TODO: support endpos in rosie
    def match(self, string, pos=1, endpos=None, **kwargs):
        if endpos is not None:
            endposIndex = endpos - 1
        else:
            endposIndex = len(string)
        encoder = kwargs['encoder'] if 'encoder' in kwargs else 'json'
        m, l, a, ttotal, tmatch = self._internal_rplx.engine.match(self._internal_rplx,
                                                        adapt23.bytes23(string[:endposIndex]),
                                                        pos,
                                                        adapt23.bytes23(encoder))
        if a and m == False: #check if matching was halted and no match found before
            raise_halt_error()
        elif m == False:
            return None
        if 'encoder' in kwargs:
            dict = {'match':m, 'leftover':l, 'abend':a}
            return dict
        else:
            match_value = json.loads(m)
            matchObject = Match(match_value, self, string, a, pos, endpos)
            return matchObject

    # TODO: support endpos in rosie
    def fullmatch(self, string, pos=1, endpos=None, **kwargs):
        if endpos is not None:
            endposIndex = endpos - 1
        else:
            endposIndex = len(string)
        m, l, a, ttotal, tmatch = self._internal_rplx.engine.match(self._internal_rplx,
                                                        adapt23.bytes23(string[:endposIndex]),
                                                        pos,
                                                        adapt23.bytes23('json'))
        if a and m == False: #check if matching was halted and no match found before
            raise_halt_error()
        elif m == False or l != 0:
            return None
        match_value = json.loads(m)
        matchObject = Match(match_value, self, string, a, pos, endpos)
        return matchObject

    # TODO: support endpos in rosie
    def trace(self, string, pos=1, endpos=None, **kwargs):
        if endpos is not None:
            endposIndex = endpos - 1
        else:
            endposIndex = len(string)
        encoder = kwargs['encoder'] if 'encoder' in kwargs else 'json'
        matched, trace_data = self._internal_rplx.engine.trace(self._internal_rplx,
                                                               adapt23.bytes23(string[:endposIndex]),
                                                               pos,
                                                               adapt23.bytes23(encoder))
        if 'encoder' in kwargs:
            return {'matched': matched, 'trace': trace_data}
        else:
            trace_value = json.loads(trace_data)
            return Trace(matched, trace_value)

    def search(self, string, pos=1, endpos=None, **kwargs):
        pattern = "find:{{({})}}".format(self.pattern)
        rplx_object = self.engine.compile(pattern)
        match_object = rplx_object.match(string, pos, endpos, **kwargs)
        if match_object is None:
            return None
        m = match_object.rosie_match
        for s in m['subs']:
            return Match(s, rplx_object, string, match_object.abend(), pos, endpos)

    def findall(self, string, pos=1, endpos=None, **kwargs):
        pattern = "findall:{{({})}}".format(self.pattern)
        rplx_object = self.engine.compile(pattern)
        match_object = rplx_object.match(string, pos, endpos, **kwargs)
        if match_object is None:
            return None
        m = match_object.rosie_match
        subs = []
        if 'subs' in m:
            for s in m['subs']:
                subs.append(s['data'])
        return subs

    def finditer(self, string, pos=1, endpos=None, **kwargs):
        pattern = "findall:{{({})}}".format(self.pattern)
        rplx_object = self.engine.compile(pattern)
        match_object = rplx_object.match(string, pos, endpos, **kwargs)
        if match_object is None:
            return None
        m = match_object.rosie_match
        subs = []
        if 'subs' in m:
            for s in m['subs']:
                subs.append(s['data'])
        return iter(subs)
    
    def sub(self, repl, string, count = 0):
        pattern = "find:{{({})}}".format(self.pattern)
        rplx_object = self.engine.compile(pattern)
        match_object = rplx_object.match(string)
        found = (match_object != None)
        while (found):
            m = match_object.rosie_match['subs'][0]
            string = string[0:m['s'] - 1] + repl + string[(m['e'] - 1):]
            match_object = rplx_object.match(string)
            found = (match_object != None)
        return string

    def subn(self, repl, string, count = 0):
        pattern = "find:{{({})}}".format(self.pattern)
        rplx_object = self.engine.compile(pattern)
        count_of_subs = 0
        match_object = rplx_object.match(string)
        found = (match_object != None)
        while (found):
            m = match_object.rosie_match['subs'][0]
            string = string[0:m['s'] - 1] + repl + string[(m['e'] - 1):]
            count_of_subs += 1
            match_object = rplx_object.match(string)
            found = (match_object != None)
        return (string, count_of_subs)



class Trace(object):
    def __init__(self, matchedValue, trace):
        self.matchedValue = matchedValue
        self.trace = trace

    def trace_value(self):
        return self.trace

    def matched(self):
        return self.matchedValue

class Match(object):

    def __init__(self, rosie_match, rplx_object, stringObject, a, position, endposition):
        self.re = rplx_object
        self.string = stringObject
        self.rosie_match = rosie_match
        self.a = a
        self.matches = []
        self.pos = position
        if (endposition == None):
            self.endpos = len(self.string)
        else:
            self.endpos = endposition

    #convert subs tree into list
    def _convertSubsToGroups(self, m):
        list = []
        match = Match(m,self.re,self.string,self.a,self.pos,self.endpos)
        list.append(match)
        if 'subs' in m:
            for s in m['subs']:
                list.extend(self._convertSubsToGroups(s))
        return list

    def expand(self, template):
        '''Return the string obtained by doing backslash substitution on the 
        template string template, as done by the sub() method. Escapes such 
        as \n are converted to the appropriate characters, and numeric 
        backreferences (\1, \2) and named backreferences (\g<1>, \g<name>) 
        are replaced by the contents of the corresponding group. Changed in 
        version 3.5: Unmatched groups are replaced with an empty string.'''
        new_string = template
        new_engine = engine()
        pattern1 = r'find:{"\\"[0-9]+}'
        pattern2 = r'find:{"\\" "g<" ([0-9]+) ">"}'
        pattern3 = r'find:{"\\" "g<" ([[0-9][a-z][A-Z]]+) ">"}'

        #pattern1
        match_object = new_engine.match(pattern1, new_string)
        found = (match_object != None)
        while (found):
            m = match_object.rosie_match['subs'][0]
            group_num = int(new_engine.search("[0-9]+", m['data']).group())
            item = self.group(group_num)
            new_string = new_string[0:m['s'] - 1] + item + new_string[(m['e'] - 1):]
            match_object = new_engine.match(pattern1, new_string)
            found = (match_object != None)

        #pattern2
        match_object = new_engine.match(pattern2, new_string)
        found = (match_object != None)
        while (found):
            m = match_object.rosie_match['subs'][0]
            group_num = int(new_engine.search("[0-9]+", m['data']).group())
            item = self.group(group_num)
            new_string = new_string[0:(m['s'] - 1)] + item + new_string[(m['e'] - 1):]
            match_object = new_engine.match(pattern2, new_string)
            found = (match_object != None)

        #pattern3
        match_object = new_engine.match(pattern3, new_string)
        found = (match_object != None)
        while (found):
            m = match_object.rosie_match['subs'][0]
            group_num = int(new_engine.search("[0-9]+", m['data']).group())
            item = self.group(group_num)
            new_string = new_string[0:(m['s'] - 1)] + item + new_string[(m['e'] - 1):]
            match_object = new_engine.match(pattern3, new_string)
            found = (match_object != None)
        return new_string


    def __getitem__(self, groupIdentifier):
        return self.group(groupIdentifier)

    def group(self, *args):
        #return original match
        if (not args):
            return self.rosie_match['data']
        else: 
            if (len(self.matches) is 0):
                self.matches.extend(self._convertSubsToGroups(self.rosie_match))
        #find capture by number
        if isinstance(args[0], int):
            if (len(args) is 1):
                m = self.matches[args[0]]
                return m.rosie_match['data']
            list = []
            for groupIndex in args:
                m = self.matches[groupIndex]
                list.append(m.rosie_match['data'])
            return tuple(list)

        #find capture by name
        #assume that isinstance(args[0], str) or isinstance(args[0], unicode) or isinstance(args[0], bytes):
        if (len(args) is 1):
            for match in self.matches:
                if match.rosie_match['type'] == args[0]:
                    return match.rosie_match['data']
            return None
        list = []
        for namedCapture in args:
            found = False
            for match in self.matches:
                if match.rosie_match['type'] == namedCapture:
                    list.append(match.rosie_match['data'])
                    found = True
            if found is False:
                list.append(None)
        return tuple(list)
            

    def groups(self):
        '''Return a tuple containing all the subgroups of the match, from 1 up to however many groups are in the pattern. 
        The default argument is used for groups that did not participate in the match; it defaults to None'''
        if (len(self.matches) is 0):
            self.matches.extend(self._convertSubsToGroups(self.rosie_match))
        list = []
        for match in self.matches:
            list.append(match.rosie_match['data'])
        del list[0]
        return tuple(list)

    def groupdict(self):
        '''Return a dictionary containing all the named subgroups of the match, keyed by the subgroup name. 
        The default argument is used for groups that did not participate in the match; it defaults to None.'''
        if (len(self.matches) is 0):
            self.matches.extend(self._convertSubsToGroups(self.rosie_match))
        dict = {}
        for match in self.matches:
            dict[match.rosie_match['type']] = match.rosie_match['data']
        return dict

    def subs(self, *args):
        if (len(self.matches) is 0):
                self.matches.extend(self._convertSubsToGroups(self.rosie_match))

        #return original match
        if (not args):
            subs = []
            for match in self.matches:
                subs.append(match.rosie_match['data'])
            return subs

        #find capture by number
        if isinstance(args[0], int):
            if (len(args) is 1):
                if (len(args[0].matches) is 0):
                    args[0].matches.extend(args[0]._convertSubsToGroups(args[0].rosie_match))
                subs = []
                for match in args[0].matches:
                    subs.append(match.rosie_match['data'])
                return subs
            else:
                raise_rosie_error("Invalid arguments: must be one argument")

        #find capture by name
        #elif isinstance(args[0], str):
        else:
            if (len(args) is 1):
                for match in self.matches:
                    if match.rosie_match['type'] == args[0]:
                        if (len(match.matches) is 0):
                            match.matches.extend(match._convertSubsToGroups(match.rosie_match))
                        subs = []
                        for submatch in match.matches:
                            subs.append(submatch.rosie_match['data'])
                        return subs
                return None

            raise_rosie_error("Invalid arguments: must be one argument")
        #else:
        #    raise_rosie_error("Invalid arguments: must be type int or string")

    def start(self, *args):
        if (not args):
            return self.rosie_match['s']
        else:
            if (len(self.matches) is 0):
                self.matches.extend(self._convertSubsToGroups(self.rosie_match))

        if isinstance(args[0], int):
            if (len(args) is 1):
                m = self.matches[args[0]]
                return m.rosie_match['s']
            else:
                raise_rosie_error("Invalid arguments: only 0-1 arguments")
        #assume isinstance(args[0], str) or isinstance(args[0], unicode) or isinstance(args[0], bytes):
        if (len(args) is 1):
            for match in self.matches:
                if match.rosie_match['type'] == args[0]:
                    return match.rosie_match['s']
        else:
            raise_rosie_error("Invalid arguments: only 0-1 arguments")

    def end(self, *args):
        if (not args):
            return self.rosie_match['e']
        else:
            if (len(self.matches) is 0):
                self.matches.extend(self._convertSubsToGroups(self.rosie_match))

        if isinstance(args[0], int):
            if (len(args) is 1):
                m = self.matches[args[0]]
                return m.rosie_match['e']
            else:
                raise_rosie_error("Invalid arguments: only 0-1 arguments")
        #assume isinstance(args[0], str):
        if (len(args) is 1):
            for match in self.matches:
                if match.rosie_match['type'] == args[0]:
                    return match.rosie_match['e']
        else:
            raise_rosie_error("Invalid arguments: only 0-1 arguments")

    def span(self, *args):
        if (not args):
            return (self.start(), self.end())
        else:
            return (self.start(*args), self.end(*args))

    def lastindex(self):
        #integer index of the last matched capturing group, or None if no group was matched at all
        if (len(self.matches) is 0):
                self.matches.extend(self._convertSubsToGroups(self.rosie_match))
        return len(self.matches) - 1

    def lastgroup(self):
        #name of the last matched capturing group, or None if the group didn’t have a name,
        #or if no group was matched at all
        if (len(self.matches) is 0):
                self.matches.extend(self._convertSubsToGroups(self.rosie_match))
        return self.matches[len(self.matches) - 1].rosie_match['type']

    def abend(self):
        return self.a

    
    







from __future__ import print_function
from setuptools import setup, find_packages
from setuptools.command.install import install
from setuptools.command.sdist import sdist
from distutils.command.build import build
from distutils.command.clean import clean
import subprocess, os, sys, shutil

ROSIE_VERSION = '1.1.0'
TAG = 'v' + ROSIE_VERSION

setup(
    name="rosie",
    version="1.1.2",

    # Do NOT set include_package_data to True, else the directory
    # contents will be included in the binary (wheel) distribution!
    include_package_data=False,

    packages = ['rosie'],

    install_requires=['cffi >= 1.9'],

    # metadata for upload to PyPI
    author="Jenna Shockley, Jamie A. Jennings",
    author_email="rosie.pattern.language@gmail.com",
    description="Rosie Pattern Language (replaces regex for data mining and text search)",

    long_description="""
Prerequisites:
    Rosie installation (see https://gitlab.com/rosie-pattern-language/rosie)

Rosie and the Rosie Pattern Language (RPL)

RPL expressions are patterns for matching text, similar to regex but
more powerful.  You can use RPL for text pattern matching the way you
might use PCRE or regex in Perl, Python, or Java.  Unlike regex, RPL
is readable and maintainable, and packages of rpl are easily shared.

The Rosie project provides a library so you can use RPL from a variety
of programming languages.  We also provide an interactive read-eval-
print loop for pattern development and debugging, and an RPL compiler.
The Rosie matching engine is very small and reasonably fast.

Rosie's home page:
  https://rosie-lang.org

The repository of record for the Rosie project is located at:
  https://gitlab.com/rosie-pattern-language/rosie

Open issues are at:
  https://gitlab.com/rosie-pattern-language/rosie/issues

Before opening an issue with a bug report or an enhancement request,
please check the current open issues.
""",

    license="MIT",
    keywords="rosie pattern language PEG regex regexp data mining text search",
    url="https://rosie-lang.org",
    project_urls={
        "Issue page": "https://gitlab.com/rosie-pattern-language/rosie/issues",
        "Documentation": "https://gitlab.com/rosie-pattern-language/rosie/tree/master/doc",
        "Source Code": "https://gitlab.com/rosie-pattern-language",
    },

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',

        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',

        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Intended Audience :: System Administrators',

        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Libraries',
        'Topic :: Text Processing :: General',
        'Topic :: Utilities',
    ],

    zip_safe=False,
)
